@@ -135,17 +135,20 @@ class Graph(BaseMainGraph):
     #self.last_numbered_iri = {}
     self.world             = world
     self.c                 = None
-    
-    self.lock              = multiprocessing.RLock()
-    self.lock_level        = 0
+    self.nb_added_triples  = 0
 
+    if read_only:
+      self.lock = multiprocessing.RLock()
+      self.acquire_write_lock = self._acquire_write_lock_read_only
+      self.release_write_lock = self._release_write_lock_read_only
+      
+    self.lock_level = 0
+    
     if initialize_db:
-      self.current_blank    = multiprocessing.Value("i", 0)
-      self.current_resource = multiprocessing.Value("i", 300) # 300 first values are reserved
-      self.prop_fts         = set()
+      self.prop_fts = set()
       
       self.execute("""CREATE TABLE store (version INTEGER, current_blank INTEGER, current_resource INTEGER)""")
-      self.execute("""INSERT INTO store VALUES (9, 0, 300)""")
+      self.execute("""INSERT INTO store VALUES (10, 0, 300)""")
       self.execute("""CREATE TABLE objs (c INTEGER, s INTEGER, p INTEGER, o INTEGER)""")
       self.execute("""CREATE TABLE datas (c INTEGER, s INTEGER, p INTEGER, o BLOB, d INTEGER)""")
       self.execute("""CREATE VIEW quads AS SELECT c,s,p,o,NULL AS d FROM objs UNION ALL SELECT c,s,p,o,d FROM datas""")
@@ -162,9 +165,11 @@ class Graph(BaseMainGraph):
 
       self.execute("""CREATE INDEX index_objs_sp ON objs(s,p)""")
       self.execute("""CREATE UNIQUE INDEX index_objs_op ON objs(o,p,c,s)""") # c is for onto.classes(), etc
+      self.execute("""CREATE INDEX index_objs_c ON objs(c)""")
       
       self.execute("""CREATE INDEX index_datas_sp ON datas(s,p)""")
       self.execute("""CREATE UNIQUE INDEX index_datas_op ON datas(o,p,c,d,s)""")
+      self.execute("""CREATE INDEX index_datas_c ON datas(c)""")
       self.indexed = True
       
       self.execute("""CREATE TABLE last_numbered_iri(prefix TEXT, i INTEGER)""")
@@ -179,13 +184,15 @@ class Graph(BaseMainGraph):
         s = "\n".join(clone.db.iterdump())
         self.db.cursor().executescript(s)
         
-      version, self.current_blank, self.current_resource = self.execute("SELECT version, current_blank, current_resource FROM store").fetchone()
-      self.current_blank    = multiprocessing.Value("i", self.current_blank)
-      self.current_resource = multiprocessing.Value("i", self.execute("SELECT MAX(storid) FROM resources").fetchone()[0])
+      #version, self.current_blank, self.current_resource = self.execute("SELECT version, current_blank, current_resource FROM store").fetchone()
+      #version, self.current_blank, drop_it = self.execute("SELECT version, current_blank, current_resource FROM store").fetchone()
+      version = self.execute("SELECT version FROM store").fetchone()[0]
+      #self.current_blank    = multiprocessing.Value("i", self.current_blank)
+      #self.current_resource = multiprocessing.Value("i", self.execute("SELECT MAX(storid) FROM resources").fetchone()[0])
       
-      if clone:
-        self.current_blank    = multiprocessing.Value("i", clone.current_blank.value)
-        self.current_resource = multiprocessing.Value("i", clone.current_resource.value)
+      #if clone:
+        #self.current_blank    = multiprocessing.Value("i", clone.current_blank.value)
+        #self.current_resource = multiprocessing.Value("i", clone.current_resource.value)
         
       if version == 1:
         print("* Owlready2 * Converting quadstore to internal format 2...", file = sys.stderr)
@@ -378,6 +385,14 @@ class Graph(BaseMainGraph):
         self.db.commit()
         version += 1
         
+      if version == 9:
+        print("* Owlready2 * Converting quadstore to internal format 10...", file = sys.stderr)
+        self.execute("""CREATE INDEX index_objs_c ON objs(c)""")
+        self.execute("""CREATE INDEX index_datas_c ON datas(c)""")
+        self.execute("""UPDATE store SET version=10""")
+        self.db.commit()
+        version += 1
+        
       self.prop_fts = { storid for (storid,) in self.execute("""SELECT storid FROM prop_fts;""") }
       
       self.analyze()
@@ -386,20 +401,25 @@ class Graph(BaseMainGraph):
     self.select_abbreviate_method()
     
   def analyze(self):
+    self.nb_added_triples = 0
+    
     if self.read_only: return
     if sqlite3.sqlite_version_info[1] < 33: return # ANALYZE sqlite_schema not supported
-    
+
     #self.db.execute("""PRAGMA cache_size = -100""") # The two following queries are * faster * with a small cache!
     #import time
     #t0 = time.perf_counter()
-    #nb_datas = self.execute("""SELECT COUNT() FROM datas INDEXED BY index_datas_sp""").fetchone()[0]
-    nb_datas = self.execute("""SELECT MAX(rowid) FROM datas""").fetchone()[0]
+    nb_datas = self.execute("""SELECT COUNT() FROM datas INDEXED BY index_datas_c""").fetchone()[0]
+    #nb_datas = self.execute("""SELECT MAX(rowid) FROM datas""").fetchone()[0] or 10
     #if nb_datas: print(nb_datas, time.perf_counter() - t0)
     #t0 = time.perf_counter()
-    #nb_objs  = self.execute("""SELECT COUNT() FROM objs INDEXED BY index_objs_sp""" ).fetchone()[0]
-    nb_objs  = self.execute("""SELECT MAX(rowid) FROM objs""" ).fetchone()[0]
+    nb_objs  = self.execute("""SELECT COUNT() FROM objs INDEXED BY index_objs_c""" ).fetchone()[0]
+    #nb_objs  = self.execute("""SELECT MAX(rowid) FROM objs""" ).fetchone()[0] or 10
     #if nb_objs: print(nb_objs, time.perf_counter() - t0)
     #self.db.execute("""PRAGMA cache_size = -200000""")
+    nb_iris  = self.execute("""SELECT MAX(storid) FROM resources""" ).fetchone()[0] or 300
+    
+    #print("ANALYZE", nb_datas, nb_objs)
     
     try:
       self.execute("""DELETE FROM sqlite_stat1""")
@@ -407,11 +427,15 @@ class Graph(BaseMainGraph):
       self.execute("""PRAGMA analysis_limit = 20""")
       self.execute("""ANALYZE""")
       self.execute("""DELETE FROM sqlite_stat1""")
+      
     self.execute("""INSERT INTO sqlite_stat1 VALUES
 ('objs', 'index_objs_op', '%s 4 3 3 1'),
 ('objs', 'index_objs_sp', '%s 3 2'),
 ('datas', 'index_datas_op', '%s 4 3 3 3 1'),
-('datas', 'index_datas_sp', '%s 3 2')""" % (nb_objs, nb_objs, nb_datas, nb_datas))
+('datas', 'index_datas_sp', '%s 3 2'),
+('resources', 'index_resources_iri', '%s 1'),
+('resources', 'resources', '%s 1')
+""" % (nb_objs, nb_objs, nb_datas, nb_datas, nb_iris, nb_iris))
     self.execute("""ANALYZE sqlite_schema""")
     
   def set_indexed(self, indexed): pass
@@ -420,28 +444,17 @@ class Graph(BaseMainGraph):
     self.db.close()
     
   def acquire_write_lock(self):
-    self.lock.acquire()
+    if not self.db.in_transaction: self.execute("BEGIN IMMEDIATE")
     self.lock_level += 1
   def release_write_lock(self):
     self.lock_level -= 1
+  def _acquire_write_lock_read_only(self):
+    self.lock.acquire()
+  def _release_write_lock_read_only(self):
     self.lock.release()
   def has_write_lock(self): return self.lock_level
   
   def select_abbreviate_method(self):
-    # nb = self.execute("SELECT count(*) FROM resources").fetchone()[0]
-    # if 0 and (nb < 100000):
-    #   iri_storid = self.execute("SELECT iri, storid FROM resources").fetchall()
-    #   self.  _abbreviate_d = dict(iri_storid)
-    #   self._unabbreviate_d = dict((storid, iri) for (iri, storid) in  iri_storid)
-    #   self._abbreviate   = self._abbreviate_dict
-    #   self._unabbreviate = self._unabbreviate_dict
-    #   self._refactor     = self._refactor_dict
-    # else:
-    #   self.  _abbreviate_d = None
-    #   self._unabbreviate_d = None
-    #   self._abbreviate   = self._abbreviate_sql
-    #   self._unabbreviate = self._unabbreviate_sql
-    #   self._refactor     = self._refactor_sql
     if self.world:
       self.world._abbreviate   = self._abbreviate
       self.world._unabbreviate = self._unabbreviate
@@ -475,53 +488,25 @@ class Graph(BaseMainGraph):
   
   def ontologies_iris(self):
     for (iri,) in self.execute("SELECT iri FROM ontologies").fetchall(): yield iri
-      
-  #def _abbreviate_sql(self, iri, create_if_missing = True):
+    
   def _abbreviate(self, iri, create_if_missing = True):
     r = self.execute("SELECT storid FROM resources WHERE iri=? LIMIT 1", (iri,)).fetchone()
     if r: return r[0]
     if create_if_missing:
-      self.current_resource.value += 1
-      storid = self.current_resource.value
+      #self.current_resource.value += 1
+      #storid = self.current_resource.value
+      storid = max(self.execute("SELECT MAX(storid)+1 FROM resources").fetchone()[0], 301) # First 300 values are reserved
+      #storid = self.execute("SELECT current_resource+1 FROM store").fetchone()[0]
+      #self.execute("UPDATE store SET current_resource=?", (storid,))
       self.execute("INSERT INTO resources VALUES (?,?)", (storid, iri))
       return storid
     
-  #def _unabbreviate_sql(self, storid):
   def _unabbreviate(self, storid):
     return self.execute("SELECT iri FROM resources WHERE storid=? LIMIT 1", (storid,)).fetchone()[0]
   
-  # def _abbreviate_dict(self, iri, create_if_missing = True):
-  #   storid = self._abbreviate_d.get(iri)
-  #   if (storid is None) and create_if_missing:
-  #     self.current_resource += 1
-  #     storid = self._abbreviate_d[iri] = self.current_resource
-  #     self._unabbreviate_d[storid] = iri
-  #     self.execute("INSERT INTO resources VALUES (?,?)", (storid, iri))
-  #   return storid
-  
-  # def _unabbreviate_dict(self, storid):
-  #   return self._unabbreviate_d[storid]
-  
   def get_storid_dict(self):
     return dict(self.execute("SELECT storid, iri FROM resources").fetchall())
   
-  # def _new_numbered_iri(self, prefix):
-  #   if prefix in self.last_numbered_iri:
-  #     i = self.last_numbered_iri[prefix] = self.last_numbered_iri[prefix] + 1
-  #     return "%s%s" % (prefix, i)
-  #   else:
-  #     cur = self.execute("SELECT iri FROM resources WHERE iri GLOB ? ORDER BY LENGTH(iri) DESC, iri DESC", ("%s*" % prefix,))
-  #     while True:
-  #       iri = cur.fetchone()
-  #       if not iri: break
-  #       num = iri[0][len(prefix):]
-  #       if num.isdigit():
-  #         self.last_numbered_iri[prefix] = i = int(num) + 1
-  #         return "%s%s" % (prefix, i)
-        
-  #   self.last_numbered_iri[prefix] = 1
-  #   return "%s1" % prefix
-  
   def _new_numbered_iri_2(self, prefix):
     cur = self.execute("SELECT iri FROM resources WHERE iri GLOB ? ORDER BY LENGTH(iri) DESC, iri DESC", ("%s*" % prefix,))
     while True:
@@ -548,20 +533,14 @@ class Graph(BaseMainGraph):
     return iri
     
   
-  #def _refactor_sql(self, storid, new_iri):
   def _refactor(self, storid, new_iri):
     self.execute("UPDATE resources SET iri=? WHERE storid=?", (new_iri, storid,))
     
-  # def _refactor_dict(self, storid, new_iri):
-  #   self.execute("UPDATE resources SET iri=? WHERE storid=?", (new_iri, storid,))
-  #   del self._abbreviate_d[self._unabbreviate_d[storid]]
-  #   self.  _abbreviate_d[new_iri] = storid
-  #   self._unabbreviate_d[storid]  = new_iri
-    
   def commit(self):
     if self.current_changes != self.db.total_changes:
       self.current_changes = self.db.total_changes
-      self.execute("UPDATE store SET current_blank=?, current_resource=?", (self.current_blank.value, self.current_resource.value))
+      #self.execute("UPDATE store SET current_blank=?, current_resource=?", (self.current_blank.value, self.current_resource.value))
+      #self.execute("UPDATE store SET current_blank=?", (self.current_blank.value,))
       self.db.commit()
 
   def context_2_user_context(self, c):
@@ -572,9 +551,12 @@ class Graph(BaseMainGraph):
     return user_c
   
   def new_blank_node(self):
-    self.current_blank.value += 1
-    return -self.current_blank.value
-  
+    blank = self.execute("SELECT current_blank+1 FROM store").fetchone()[0]
+    self.execute("UPDATE store SET current_blank=?", (blank,))
+    return -blank
+    #self.current_blank.value += 1
+    #return -self.current_blank.value
+    
   def _get_obj_triples_spo_spo(self, s, p, o):
     if s is None:
       if p is None:
@@ -947,15 +929,9 @@ SELECT c, o FROM objs q1 WHERE s=? AND o < 0 AND (SELECT COUNT() FROM objs q2 WH
   
   def restore_iri(self, storid, iri):
     self.execute("INSERT INTO resources VALUES (?,?)", (storid, iri))
-    #if self._abbreviate_d:
-    #  self._unabbreviate_d[storid] = iri
-    #  self._abbreviate_d  [iri]    = storid
       
   def destroy_entity(self, storid, destroyer, relation_updater, undoer_objs = None, undoer_datas = None):
     self.execute("DELETE FROM resources WHERE storid=?", (storid,))
-    #if self._abbreviate_d and (storid > 0):
-    #  del self._abbreviate_d  [self._unabbreviate_d[storid]]
-    #  del self._unabbreviate_d[storid]
       
     destroyed_storids   = { storid }
     modified_relations  = defaultdict(set)
@@ -1101,42 +1077,22 @@ class SubGraph(BaseSubGraph):
       cur.execute("DELETE FROM objs WHERE c=?", (self.c,))
       cur.execute("DELETE FROM datas WHERE c=?", (self.c,))
       
-    #if cur.execute("""SELECT COUNT() FROM ontologies""").fetchone()[0] < 2:
-    #  cur.execute("""DROP INDEX IF EXISTS index_resources_iri""")
-    #  cur.execute("""DROP INDEX IF EXISTS index_objs_sp""")
-    #  cur.execute("""DROP INDEX IF EXISTS index_objs_op""")
-    #  cur.execute("""DROP INDEX IF EXISTS index_datas_sp""")
-    #  cur.execute("""DROP INDEX IF EXISTS index_datas_op""")
-    #  reindex = True
-    #else:
-    #  reindex = False
-    
     # Re-implement _abbreviate() for speed
-    # if self.parent._abbreviate_d is None:
     abbrevs = {}
+    current_resource = max(self.execute("SELECT MAX(storid) FROM resources").fetchone()[0], 300) # First 300 values are reserved
     def _abbreviate(iri):
+        nonlocal current_resource
         storid = abbrevs.get(iri)
         if not storid is None: return storid
         r = cur.execute("SELECT storid FROM resources WHERE iri=? LIMIT 1", (iri,)).fetchone()
         if r:
           abbrevs[iri] = r[0]
           return r[0]
-        self.parent.current_resource.value += 1
-        storid = self.parent.current_resource.value
+        current_resource += 1
+        storid = current_resource
         new_abbrevs.append((storid, iri))
         abbrevs[iri] = storid
         return storid
-    # else:
-    #   abbrevs = self.parent._abbreviate_d
-    #   def _abbreviate(iri):
-    #     storid = abbrevs.get(iri)
-    #     if not storid is None: return storid
-        
-    #     self.parent.current_resource += 1
-    #     storid = self.parent.current_resource
-    #     new_abbrevs.append((storid, iri))
-    #     abbrevs[iri] = storid
-    #     return storid
       
     def insert_objs():
       nonlocal objs, new_abbrevs
@@ -1172,15 +1128,6 @@ class SubGraph(BaseSubGraph):
       insert_objs()
       insert_datas()
       
-      #if reindex:
-      #  cur.execute("""CREATE UNIQUE INDEX index_resources_iri ON resources(iri)""")
-      #  if self.parent.indexed:
-      #    cur.execute("""CREATE INDEX index_objs_sp ON objs(s,p)""")
-      #    cur.execute("""CREATE INDEX index_objs_op ON objs(o,p)""")
-      #    cur.execute("""CREATE INDEX index_datas_sp ON datas(s,p)""")
-      #    cur.execute("""CREATE INDEX index_datas_op ON datas(o,p)""")
-      
-      t0 = time.time()
       onto_base_iri = cur.execute("SELECT resources.iri FROM objs, resources WHERE objs.c=? AND objs.o=? AND resources.storid=objs.s LIMIT 1", (self.c, owl_ontology)).fetchone()
       if onto_base_iri: onto_base_iri = onto_base_iri[0]
       else:             onto_base_iri = ""
@@ -1222,10 +1169,14 @@ class SubGraph(BaseSubGraph):
     if (s is None) or (p is None) or (o is None): raise ValueError
     self.execute("DELETE FROM objs WHERE c=? AND s=? AND p=?", (self.c, s, p,))
     self.execute("INSERT INTO objs VALUES (?, ?, ?, ?)", (self.c, s, p, o))
+    self.parent.nb_added_triples += 1
+    if self.parent.nb_added_triples > 1000: self.parent.analyze()
     
   def _add_obj_triple_raw_spo(self, s, p, o):
     if (s is None) or (p is None) or (o is None): raise ValueError
     self.execute("INSERT OR IGNORE INTO objs VALUES (?, ?, ?, ?)", (self.c, s, p, o))
+    self.parent.nb_added_triples += 1
+    if self.parent.nb_added_triples > 1000: self.parent.analyze()
     
   def _del_obj_triple_raw_spo(self, s = None, p = None, o = None):
     if s is None:
@@ -1247,10 +1198,14 @@ class SubGraph(BaseSubGraph):
     if (s is None) or (p is None) or (o is None) or (d is None): raise ValueError
     self.execute("DELETE FROM datas WHERE c=? AND s=? AND p=?", (self.c, s, p,))
     self.execute("INSERT INTO datas VALUES (?, ?, ?, ?, ?)", (self.c, s, p, o, d))
-    
+    self.parent.nb_added_triples += 1
+    if self.parent.nb_added_triples > 1000: self.parent.analyze()
+
   def _add_data_triple_raw_spod(self, s, p, o, d):
     if (s is None) or (p is None) or (o is None) or (d is None): raise ValueError
     self.execute("INSERT OR IGNORE INTO datas VALUES (?, ?, ?, ?, ?)", (self.c, s, p, o, d))
+    self.parent.nb_added_triples += 1
+    if self.parent.nb_added_triples > 1000: self.parent.analyze()
     
   def _del_data_triple_raw_spod(self, s, p, o, d):
     if s is None:
